@@ -1,6 +1,6 @@
 /*
  * HEIF codec.
- * Copyright (c) 2017 struktur AG, Dirk Farin <farin@struktur.de>
+ * Copyright (c) 2017 Dirk Farin <dirk.farin@gmail.com>
  *
  * This file is part of libheif.
  *
@@ -18,22 +18,23 @@
  * along with libheif.  If not, see <http://www.gnu.org/licenses/>.
  */
 
-#if defined(HAVE_CONFIG_H)
-#include "config.h"
-#endif
-
+#include "libheif/heif_plugin.h"
+#include "libheif/region.h"
+#include "libheif/common_utils.h"
+#include <cstdint>
 #include "heif.h"
-#include "heif_file.h"
-#include "heif_image.h"
-#include "heif_api_structs.h"
-#include "heif_context.h"
-#include "heif_plugin_registry.h"
+#include "file.h"
+#include "pixelimage.h"
+#include "api_structs.h"
+#include "context.h"
+#include "plugin_registry.h"
 #include "error.h"
 #include "bitstream.h"
+#include "init.h"
 #include <set>
 #include <limits>
 
-#if defined(__EMSCRIPTEN__)
+#if defined(__EMSCRIPTEN__) && !defined(__EMSCRIPTEN_STANDALONE_WASM__)
 #include "heif_emscripten.h"
 #endif
 
@@ -46,7 +47,7 @@
 #include <vector>
 #include <cstring>
 
-#if (defined(__MINGW32__) || defined(__MINGW64__) || defined(_MSC_VER)) && !defined(HAVE_UNISTD_H)
+#ifdef _WIN32
 // for _write
 #include <io.h>
 #else
@@ -55,9 +56,10 @@
 
 #endif
 
-using namespace heif;
+#include <cassert>
 
-static struct heif_error error_Ok = {heif_error_Ok, heif_suberror_Unspecified, kSuccess};
+
+const struct heif_error heif_error_success = {heif_error_Ok, heif_suberror_Unspecified, kSuccess};
 static struct heif_error error_unsupported_parameter = {heif_error_Usage_error,
                                                         heif_suberror_Unsupported_parameter,
                                                         "Unsupported encoder parameter"};
@@ -83,7 +85,7 @@ uint32_t heif_get_version_number(void)
 
 static uint8_t bcd2dec(uint8_t v)
 {
-  return uint8_t((v>>4) * 10 + (v & 0x0F));
+  return uint8_t((v >> 4) * 10 + (v & 0x0F));
 }
 
 int heif_get_version_number_major(void)
@@ -116,21 +118,27 @@ heif_filetype_result heif_check_filetype(const uint8_t* data, int len)
   }
 
   if (len >= 12) {
-    heif_brand brand = heif_main_brand(data, len);
+    heif_brand2 brand = heif_read_main_brand(data, len);
 
-    if (brand == heif_heic) {
+    if (brand == heif_brand2_heic) {
       return heif_filetype_yes_supported;
     }
-    else if (brand == heif_heix) {
+    else if (brand == heif_brand2_heix) {
       return heif_filetype_yes_supported;
     }
-    else if (brand == heif_avif) {
+    else if (brand == heif_brand2_avif) {
       return heif_filetype_yes_supported;
     }
-    else if (brand == heif_unknown_brand) {
-      return heif_filetype_no;
+    else if (brand == heif_brand2_jpeg) {
+      return heif_filetype_yes_supported;
     }
-    else if (brand == heif_mif1) {
+    else if (brand == heif_brand2_j2ki) {
+      return heif_filetype_yes_supported;
+    }
+    else if (brand == heif_brand2_mif1) {
+      return heif_filetype_maybe;
+    }
+    else if (brand == heif_brand2_mif2) {
       return heif_filetype_maybe;
     }
     else {
@@ -144,7 +152,7 @@ heif_filetype_result heif_check_filetype(const uint8_t* data, int len)
 
 int heif_check_jpeg_filetype(const uint8_t* data, int len)
 {
-  if (len<12 || data==nullptr) {
+  if (len < 12 || data == nullptr) {
     return -1;
   }
 
@@ -153,13 +161,13 @@ int heif_check_jpeg_filetype(const uint8_t* data, int len)
       0x4A, 0x46, 0x49, 0x46, 0x00, 0x01
   };
 
-  return strncmp((const char*)data, (const char*)jpeg_signature, 12)==0;
+  return strncmp((const char*) data, (const char*) jpeg_signature, 12) == 0;
 }
 
 
 heif_brand heif_fourcc_to_brand_enum(const char* fourcc)
 {
-  if (fourcc==nullptr || !fourcc[0] || !fourcc[1] || !fourcc[2] || !fourcc[3]) {
+  if (fourcc == nullptr || !fourcc[0] || !fourcc[1] || !fourcc[2] || !fourcc[3]) {
     return heif_unknown_brand;
   }
 
@@ -209,6 +217,12 @@ heif_brand heif_fourcc_to_brand_enum(const char* fourcc)
   else if (strcmp(brand, "vvic") == 0) {
     return heif_vvic;
   }
+  else if (strcmp(brand, "j2ki") == 0) {
+    return heif_j2ki;
+  }
+  else if (strcmp(brand, "j2is") == 0) {
+    return heif_j2is;
+  }
   else {
     return heif_unknown_brand;
   }
@@ -221,7 +235,7 @@ enum heif_brand heif_main_brand(const uint8_t* data, int len)
     return heif_unknown_brand;
   }
 
-  return heif_fourcc_to_brand_enum((char*)(data+8));
+  return heif_fourcc_to_brand_enum((char*) (data + 8));
 }
 
 
@@ -231,15 +245,13 @@ heif_brand2 heif_read_main_brand(const uint8_t* data, int len)
     return heif_unknown_brand;
   }
 
-  return heif_fourcc_to_brand((char*)(data+8));
+  return heif_fourcc_to_brand((char*) (data + 8));
 }
 
 
-#define fourcc_to_uint32(id) (((uint32_t)(id[0])<<24) | (id[1]<<16) | (id[2]<<8) | (id[3]))
-
 heif_brand2 heif_fourcc_to_brand(const char* fourcc)
 {
-  if (fourcc==nullptr || !fourcc[0] || !fourcc[1] || !fourcc[2] || !fourcc[3]) {
+  if (fourcc == nullptr || !fourcc[0] || !fourcc[1] || !fourcc[2] || !fourcc[3]) {
     return 0;
   }
 
@@ -250,24 +262,24 @@ heif_brand2 heif_fourcc_to_brand(const char* fourcc)
 void heif_brand_to_fourcc(heif_brand2 brand, char* out_fourcc)
 {
   if (out_fourcc) {
-    out_fourcc[0] = (char)((brand >> 24) & 0xFF);
-    out_fourcc[1] = (char)((brand >> 16) & 0xFF);
-    out_fourcc[2] = (char)((brand >>  8) & 0xFF);
-    out_fourcc[3] = (char)((brand >>  0) & 0xFF);
+    out_fourcc[0] = (char) ((brand >> 24) & 0xFF);
+    out_fourcc[1] = (char) ((brand >> 16) & 0xFF);
+    out_fourcc[2] = (char) ((brand >> 8) & 0xFF);
+    out_fourcc[3] = (char) ((brand >> 0) & 0xFF);
   }
 }
 
 
 int heif_has_compatible_brand(const uint8_t* data, int len, const char* brand_fourcc)
 {
-  if (data == nullptr || len<=0 || brand_fourcc == nullptr || !brand_fourcc[0] || !brand_fourcc[1] || !brand_fourcc[2] || !brand_fourcc[3]) {
+  if (data == nullptr || len <= 0 || brand_fourcc == nullptr || !brand_fourcc[0] || !brand_fourcc[1] || !brand_fourcc[2] || !brand_fourcc[3]) {
     return -1;
   }
 
   auto stream = std::make_shared<StreamReader_memory>(data, len, false);
   BitstreamRange range(stream, len);
 
-  std::shared_ptr<heif::Box> box;
+  std::shared_ptr<Box> box;
   Error err = Box::read(range, &box);
   if (err) {
     if (err.sub_error_code == heif_suberror_End_of_data) {
@@ -286,21 +298,20 @@ int heif_has_compatible_brand(const uint8_t* data, int len, const char* brand_fo
 }
 
 
-
 struct heif_error heif_list_compatible_brands(const uint8_t* data, int len, heif_brand2** out_brands, int* out_size)
 {
-  if (data == nullptr || out_brands==nullptr || out_size==nullptr) {
+  if (data == nullptr || out_brands == nullptr || out_size == nullptr) {
     return {heif_error_Usage_error, heif_suberror_Null_pointer_argument, "NULL argument"};
   }
 
-  if (len<=0) {
+  if (len <= 0) {
     return {heif_error_Usage_error, heif_suberror_Invalid_parameter_value, "data length must be positive"};
   }
 
   auto stream = std::make_shared<StreamReader_memory>(data, len, false);
   BitstreamRange range(stream, len);
 
-  std::shared_ptr<heif::Box> box;
+  std::shared_ptr<Box> box;
   Error err = Box::read(range, &box);
   if (err) {
     if (err.sub_error_code == heif_suberror_End_of_data) {
@@ -316,14 +327,14 @@ struct heif_error heif_list_compatible_brands(const uint8_t* data, int len, heif
   }
 
   auto brands = ftyp->list_brands();
-  *out_brands = (heif_brand2*)malloc(sizeof(heif_brand2) * brands.size());
-  *out_size = (int)brands.size();
+  *out_brands = (heif_brand2*) malloc(sizeof(heif_brand2) * brands.size());
+  *out_size = (int) brands.size();
 
-  for (int i=0;i<(int)brands.size();i++) {
+  for (int i = 0; i < (int) brands.size(); i++) {
     (*out_brands)[i] = brands[i];
   }
 
-  return {heif_error_Ok, heif_suberror_Unspecified, Error::kSuccess};
+  return heif_error_success;
 }
 
 
@@ -406,6 +417,12 @@ const char* heif_get_file_mime_type(const uint8_t* data, int len)
   else if (mainBrand == heif_avis) {
     return "image/avif-sequence";
   }
+  else if (mainBrand == heif_j2ki) {
+    return "image/hej2k";
+  }
+  else if (mainBrand == heif_j2is) {
+    return "image/j2is";
+  }
   else if (is_jpeg(data, len) == TriBool::Yes) {
     return "image/jpeg";
   }
@@ -420,6 +437,8 @@ const char* heif_get_file_mime_type(const uint8_t* data, int len)
 
 heif_context* heif_context_alloc()
 {
+  load_plugins_if_not_initialized_yet();
+
   struct heif_context* ctx = new heif_context;
   ctx->context = std::make_shared<HeifContext>();
 
@@ -473,8 +492,8 @@ void heif_context_debug_dump_boxes_to_file(struct heif_context* ctx, int fd)
 
   std::string dump = ctx->context->debug_dump_boxes();
   // TODO(fancycode): Should we return an error if writing fails?
-#if (defined(__MINGW32__) || defined(__MINGW64__) || defined(_MSC_VER)) && !defined(HAVE_UNISTD_H)
-  auto written = _write(fd, dump.c_str(), dump.size());
+#ifdef _WIN32
+  auto written = _write(fd, dump.c_str(), static_cast<unsigned int>(dump.size()));
 #else
   auto written = write(fd, dump.c_str(), dump.size());
 #endif
@@ -572,31 +591,22 @@ struct heif_error heif_context_get_image_handle(struct heif_context* ctx,
                                                 struct heif_image_handle** imgHdl)
 {
   if (!imgHdl) {
-    Error err(heif_error_Usage_error,
-              heif_suberror_Null_pointer_argument);
-    return err.error_struct(ctx->context.get());
+    return {heif_error_Usage_error, heif_suberror_Null_pointer_argument, ""};
   }
 
-  const std::vector<std::shared_ptr<HeifContext::Image>> images = ctx->context->get_top_level_images();
-
-  std::shared_ptr<HeifContext::Image> image;
-  for (auto& img : images) {
-    if (img->get_id() == id) {
-      image = img;
-      break;
-    }
-  }
+  auto image = ctx->context->get_image(id);
 
   if (!image) {
-    Error err(heif_error_Usage_error, heif_suberror_Nonexisting_item_referenced);
-    return err.error_struct(ctx->context.get());
+    *imgHdl = nullptr;
+
+    return {heif_error_Usage_error, heif_suberror_Nonexisting_item_referenced, ""};
   }
 
   *imgHdl = new heif_image_handle();
   (*imgHdl)->image = image;
   (*imgHdl)->context = ctx->context;
 
-  return Error::Ok.error_struct(ctx->context.get());
+  return heif_error_success;
 }
 
 
@@ -606,6 +616,12 @@ int heif_image_handle_is_primary_image(const struct heif_image_handle* handle)
 }
 
 
+heif_item_id heif_image_handle_get_item_id(const struct heif_image_handle* handle)
+{
+  return handle->image->get_id();
+}
+
+
 int heif_image_handle_get_number_of_thumbnails(const struct heif_image_handle* handle)
 {
   return (int) handle->image->get_thumbnails().size();
@@ -691,7 +707,7 @@ struct heif_error heif_image_handle_get_auxiliary_type(const struct heif_image_h
 
   auto auxType = handle->image->get_aux_type();
 
-  char* buf = (char*)malloc(auxType.length()+1);
+  char* buf = (char*) malloc(auxType.length() + 1);
 
   if (buf == nullptr) {
     return Error(heif_error_Memory_allocation_error,
@@ -702,17 +718,24 @@ struct heif_error heif_image_handle_get_auxiliary_type(const struct heif_image_h
   strcpy(buf, auxType.c_str());
   *out_type = buf;
 
-  return {heif_error_Ok, heif_suberror_Unspecified, Error::kSuccess};
+  return heif_error_success;
 }
 
 
+void heif_image_handle_release_auxiliary_type(const struct heif_image_handle* handle,
+                                              const char** out_type)
+{
+  if (out_type && *out_type) {
+    free((void*) *out_type);
+    *out_type = nullptr;
+  }
+}
+
+// DEPRECATED (typo)
 void heif_image_handle_free_auxiliary_types(const struct heif_image_handle* handle,
                                             const char** out_type)
 {
-  if (out_type && *out_type) {
-    free((void*)*out_type);
-    *out_type = nullptr;
-  }
+  heif_image_handle_release_auxiliary_type(handle, out_type);
 }
 
 
@@ -785,6 +808,27 @@ int heif_image_handle_get_ispe_height(const struct heif_image_handle* handle)
 }
 
 
+struct heif_context* heif_image_handle_get_context(const struct heif_image_handle* handle)
+{
+  auto ctx = new heif_context();
+  ctx->context = handle->context;
+  return ctx;
+}
+
+
+struct heif_error heif_image_handle_get_preferred_decoding_colorspace(const struct heif_image_handle* image_handle,
+                                                                      enum heif_colorspace* out_colorspace,
+                                                                      enum heif_chroma* out_chroma)
+{
+  Error err = image_handle->image->get_preferred_decoding_colorspace(out_colorspace, out_chroma);
+  if (err) {
+    return err.error_struct(image_handle->image.get());
+  }
+
+  return heif_error_success;
+}
+
+
 int heif_image_handle_has_alpha_channel(const struct heif_image_handle* handle)
 {
   // TODO: for now, also scan the grid tiles for alpha information (issue #708), but depending about
@@ -907,30 +951,69 @@ struct heif_error heif_image_handle_get_depth_image_handle(const struct heif_ima
 }
 
 
+void fill_default_decoding_options(heif_decoding_options& options)
+{
+  options.version = 5;
+
+  options.ignore_transformations = false;
+
+  options.start_progress = nullptr;
+  options.on_progress = nullptr;
+  options.end_progress = nullptr;
+  options.progress_user_data = nullptr;
+
+  // version 2
+
+  options.convert_hdr_to_8bit = false;
+
+  // version 3
+
+  options.strict_decoding = false;
+
+  // version 4
+
+  options.decoder_id = nullptr;
+
+  // version 5
+
+  options.color_conversion_options.version = 1;
+  options.color_conversion_options.preferred_chroma_downsampling_algorithm = heif_chroma_downsampling_average;
+  options.color_conversion_options.preferred_chroma_upsampling_algorithm = heif_chroma_upsampling_bilinear;
+  options.color_conversion_options.only_use_preferred_chroma_algorithm = false;
+}
+
+
+static void copy_options(heif_decoding_options& options, const heif_decoding_options& input_options)
+{
+  switch (input_options.version) {
+    case 5:
+      options.color_conversion_options = input_options.color_conversion_options;
+      // fallthrough
+    case 4:
+      options.decoder_id = input_options.decoder_id;
+      // fallthrough
+    case 3:
+      options.strict_decoding = input_options.strict_decoding;
+      // fallthrough
+    case 2:
+      options.convert_hdr_to_8bit = input_options.convert_hdr_to_8bit;
+      // fallthrough
+    case 1:
+      options.ignore_transformations = input_options.ignore_transformations;
+
+      options.start_progress = input_options.start_progress;
+      options.on_progress = input_options.on_progress;
+      options.end_progress = input_options.end_progress;
+      options.progress_user_data = input_options.progress_user_data;
+  }
+}
+
+
 heif_decoding_options* heif_decoding_options_alloc()
 {
   auto options = new heif_decoding_options;
 
-  options->version = 4;
-
-  options->ignore_transformations = false;
-
-  options->start_progress = nullptr;
-  options->on_progress = nullptr;
-  options->end_progress = nullptr;
-  options->progress_user_data = nullptr;
-
-  // version 2
-
-  options->convert_hdr_to_8bit = false;
-
-  // version 3
-
-  options->strict_decoding = false;
-
-  // version 4
-
-  options->decoder_id = nullptr;
+  fill_default_decoding_options(*options);
 
   return options;
 }
@@ -946,16 +1029,24 @@ struct heif_error heif_decode_image(const struct heif_image_handle* in_handle,
                                     struct heif_image** out_img,
                                     heif_colorspace colorspace,
                                     heif_chroma chroma,
-                                    const struct heif_decoding_options* options)
+                                    const struct heif_decoding_options* input_options)
 {
   std::shared_ptr<HeifPixelImage> img;
 
   heif_item_id id = in_handle->image->get_id();
 
+  heif_decoding_options dec_options;
+  fill_default_decoding_options(dec_options);
+
+  if (input_options != nullptr) {
+    // overwrite the (possibly lower version) input options over the default options
+    copy_options(dec_options, *input_options);
+  }
+
   Error err = in_handle->context->decode_image_user(id, img,
                                                     colorspace,
                                                     chroma,
-                                                    options);
+                                                    dec_options);
   if (err.error_code != heif_error_Ok) {
     return err.error_struct(in_handle->image.get());
   }
@@ -972,6 +1063,25 @@ struct heif_error heif_image_create(int width, int height,
                                     heif_chroma chroma,
                                     struct heif_image** image)
 {
+  if (image == nullptr) {
+    return {heif_error_Usage_error, heif_suberror_Null_pointer_argument, "heif_image_create: NULL passed as image pointer."};
+  }
+
+  // auto-correct colorspace_YCbCr + chroma_monochrome to colorspace_monochrome
+  // TODO: this should return an error in a later version (see below)
+  if (chroma == heif_chroma_monochrome && colorspace == heif_colorspace_YCbCr) {
+    colorspace = heif_colorspace_monochrome;
+
+    std::cerr << "libheif warning: heif_image_create() used with an illegal colorspace/chroma combination. This will return an error in a future version.\n";
+  }
+
+  // return error if invalid colorspace + chroma combination is used
+  auto validChroma = get_valid_chroma_values_for_colorspace(colorspace);
+  if (std::find(validChroma.begin(), validChroma.end(), chroma) == validChroma.end()) {
+    *image = nullptr;
+    return {heif_error_Usage_error, heif_suberror_Invalid_parameter_value, "Invalid colorspace/chroma combination."};
+  }
+
   struct heif_image* img = new heif_image;
   img->image = std::make_shared<HeifPixelImage>();
 
@@ -979,8 +1089,7 @@ struct heif_error heif_image_create(int width, int height,
 
   *image = img;
 
-  struct heif_error err = {heif_error_Ok, heif_suberror_Unspecified, Error::kSuccess};
-  return err;
+  return heif_error_success;
 }
 
 int heif_image_get_decoding_warnings(struct heif_image* image,
@@ -1022,7 +1131,7 @@ void heif_image_get_content_light_level(const struct heif_image* image, struct h
 
 void heif_image_set_content_light_level(const struct heif_image* image, const struct heif_content_light_level* in)
 {
-  if (in==nullptr) {
+  if (in == nullptr) {
     return;
   }
 
@@ -1042,7 +1151,7 @@ void heif_image_get_mastering_display_colour_volume(const struct heif_image* ima
 
 void heif_image_set_mastering_display_colour_volume(const struct heif_image* image, const struct heif_mastering_display_colour_volume* in)
 {
-  if (in==nullptr) {
+  if (in == nullptr) {
     return;
   }
 
@@ -1052,31 +1161,31 @@ void heif_image_set_mastering_display_colour_volume(const struct heif_image* ima
 float mdcv_coord_decode_x(uint16_t coord)
 {
   // check for unspecified value
-  if (coord<5 || coord>37000) {
+  if (coord < 5 || coord > 37000) {
     return 0.0f;
   }
 
-  return (float)(coord * 0.00002);
+  return (float) (coord * 0.00002);
 }
 
 float mdcv_coord_decode_y(uint16_t coord)
 {
   // check for unspecified value
-  if (coord<5 || coord>42000) {
+  if (coord < 5 || coord > 42000) {
     return 0.0f;
   }
 
-  return (float)(coord * 0.00002);
+  return (float) (coord * 0.00002);
 }
 
 struct heif_error heif_mastering_display_colour_volume_decode(const struct heif_mastering_display_colour_volume* in,
                                                               struct heif_decoded_mastering_display_colour_volume* out)
 {
-  if (in==nullptr || out==nullptr) {
+  if (in == nullptr || out == nullptr) {
     return error_null_parameter;
   }
 
-  for (int c=0;c<3;c++) {
+  for (int c = 0; c < 3; c++) {
     out->display_primaries_x[c] = mdcv_coord_decode_x(in->display_primaries_x[c]);
     out->display_primaries_y[c] = mdcv_coord_decode_y(in->display_primaries_y[c]);
   }
@@ -1098,7 +1207,7 @@ struct heif_error heif_mastering_display_colour_volume_decode(const struct heif_
     out->min_display_mastering_luminance = in->min_display_mastering_luminance * 0.0001;
   }
 
-  return error_Ok;
+  return heif_error_success;
 }
 
 
@@ -1194,7 +1303,7 @@ heif_error heif_image_crop(struct heif_image* img,
 
   img->image = out_img;
 
-  return heif_error{heif_error_Ok, heif_suberror_Unspecified, Error::kSuccess};
+  return heif_error_success;
 }
 
 
@@ -1226,8 +1335,7 @@ struct heif_error heif_image_add_plane(struct heif_image* image,
     return err;
   }
   else {
-    struct heif_error err = {heif_error_Ok, heif_suberror_Unspecified, Error::kSuccess};
-    return err;
+    return heif_error_success;
   }
 }
 
@@ -1288,7 +1396,7 @@ struct heif_error heif_image_extend_padding_to_size(struct heif_image* image, in
                       "Cannot allocate image memory."};
   }
   else {
-    return heif_error{heif_error_Ok, heif_suberror_Unspecified, Error::kSuccess};
+    return heif_error_success;
   }
 }
 
@@ -1334,8 +1442,7 @@ struct heif_error heif_image_set_raw_color_profile(struct heif_image* image,
 
   image->image->set_color_profile_icc(color_profile);
 
-  struct heif_error err = {heif_error_Ok, heif_suberror_Unspecified, Error::kSuccess};
-  return err;
+  return heif_error_success;
 }
 
 
@@ -1351,7 +1458,7 @@ struct heif_error heif_image_set_nclx_color_profile(struct heif_image* image,
 
   image->image->set_color_profile_nclx(nclx);
 
-  return error_Ok;
+  return heif_error_success;
 }
 
 
@@ -1426,6 +1533,19 @@ const char* heif_image_handle_get_metadata_content_type(const struct heif_image_
 }
 
 
+const char* heif_image_handle_get_metadata_item_uri_type(const struct heif_image_handle* handle,
+                                                         heif_item_id metadata_id)
+{
+  for (auto& metadata : handle->image->get_metadata()) {
+    if (metadata->item_id == metadata_id) {
+      return metadata->item_uri_type.c_str();
+    }
+  }
+
+  return nullptr;
+}
+
+
 size_t heif_image_handle_get_metadata_size(const struct heif_image_handle* handle,
                                            heif_item_id metadata_id)
 {
@@ -1531,23 +1651,23 @@ struct heif_error heif_nclx_color_profile_set_color_primaries(heif_color_profile
 
 
 static const std::set<typename std::underlying_type<heif_transfer_characteristics>::type> known_transfer_characteristics{
-  heif_transfer_characteristic_ITU_R_BT_709_5,
-  heif_transfer_characteristic_unspecified,
-  heif_transfer_characteristic_ITU_R_BT_470_6_System_M,
-  heif_transfer_characteristic_ITU_R_BT_470_6_System_B_G,
-  heif_transfer_characteristic_ITU_R_BT_601_6,
-  heif_transfer_characteristic_SMPTE_240M,
-  heif_transfer_characteristic_linear,
-  heif_transfer_characteristic_logarithmic_100,
-  heif_transfer_characteristic_logarithmic_100_sqrt10,
-  heif_transfer_characteristic_IEC_61966_2_4,
-  heif_transfer_characteristic_ITU_R_BT_1361,
-  heif_transfer_characteristic_IEC_61966_2_1,
-  heif_transfer_characteristic_ITU_R_BT_2020_2_10bit,
-  heif_transfer_characteristic_ITU_R_BT_2020_2_12bit,
-  heif_transfer_characteristic_ITU_R_BT_2100_0_PQ,
-  heif_transfer_characteristic_SMPTE_ST_428_1,
-  heif_transfer_characteristic_ITU_R_BT_2100_0_HLG
+    heif_transfer_characteristic_ITU_R_BT_709_5,
+    heif_transfer_characteristic_unspecified,
+    heif_transfer_characteristic_ITU_R_BT_470_6_System_M,
+    heif_transfer_characteristic_ITU_R_BT_470_6_System_B_G,
+    heif_transfer_characteristic_ITU_R_BT_601_6,
+    heif_transfer_characteristic_SMPTE_240M,
+    heif_transfer_characteristic_linear,
+    heif_transfer_characteristic_logarithmic_100,
+    heif_transfer_characteristic_logarithmic_100_sqrt10,
+    heif_transfer_characteristic_IEC_61966_2_4,
+    heif_transfer_characteristic_ITU_R_BT_1361,
+    heif_transfer_characteristic_IEC_61966_2_1,
+    heif_transfer_characteristic_ITU_R_BT_2020_2_10bit,
+    heif_transfer_characteristic_ITU_R_BT_2020_2_12bit,
+    heif_transfer_characteristic_ITU_R_BT_2100_0_PQ,
+    heif_transfer_characteristic_SMPTE_ST_428_1,
+    heif_transfer_characteristic_ITU_R_BT_2100_0_HLG
 };
 
 
@@ -1739,19 +1859,10 @@ void heif_nclx_color_profile_free(struct heif_color_profile_nclx* nclx_profile)
   color_profile_nclx::free_nclx_color_profile(nclx_profile);
 }
 
-
 // DEPRECATED
 struct heif_error heif_register_decoder(heif_context* heif, const heif_decoder_plugin* decoder_plugin)
 {
-  if (!decoder_plugin) {
-    return error_null_parameter;
-  }
-  else if (decoder_plugin->plugin_api_version > 3) {
-    return error_unsupported_plugin_version;
-  }
-
-  heif->context->register_decoder(decoder_plugin);
-  return Error::Ok.error_struct(heif->context.get());
+  return heif_register_decoder_plugin(decoder_plugin);
 }
 
 
@@ -1765,7 +1876,7 @@ struct heif_error heif_register_decoder_plugin(const heif_decoder_plugin* decode
   }
 
   register_decoder(decoder_plugin);
-  return error_Ok;
+  return heif_error_success;
 }
 
 struct heif_error heif_register_encoder_plugin(const heif_encoder_plugin* encoder_plugin)
@@ -1778,7 +1889,7 @@ struct heif_error heif_register_encoder_plugin(const heif_encoder_plugin* encode
   }
 
   register_encoder(encoder_plugin);
-  return error_Ok;
+  return heif_error_success;
 }
 
 
@@ -1843,7 +1954,13 @@ struct heif_error heif_context_write(struct heif_context* ctx,
   ctx->context->write(swriter);
 
   const auto& data = swriter.get_data();
-  return writer->write(ctx, data.data(), data.size(), userdata);
+  heif_error writer_error = writer->write(ctx, data.data(), data.size(), userdata);
+  if (!writer_error.message) {
+    return heif_error{heif_error_Usage_error, heif_suberror_Null_pointer_argument, "heif_writer callback returned a null error text"};
+  }
+  else {
+    return writer_error;
+  }
 }
 
 
@@ -1862,13 +1979,17 @@ int heif_get_encoder_descriptors(enum heif_compression_format format,
                                  const struct heif_encoder_descriptor** out_encoder_descriptors,
                                  int count)
 {
-  if (out_encoder_descriptors == nullptr || count <= 0) {
+  if (out_encoder_descriptors != nullptr && count <= 0) {
     return 0;
   }
 
   std::vector<const struct heif_encoder_descriptor*> descriptors;
   descriptors = get_filtered_encoder_descriptors(format, name);
 
+  if (out_encoder_descriptors == nullptr) {
+    return static_cast<int>(descriptors.size());
+  }
+
   int i;
   for (i = 0; i < count && static_cast<size_t>(i) < descriptors.size(); i++) {
     out_encoder_descriptors[i] = descriptors[i];
@@ -1894,7 +2015,8 @@ int heif_get_decoder_descriptors(enum heif_compression_format format_filter,
                                  const struct heif_decoder_descriptor** out_decoders,
                                  int count)
 {
-  struct decoder_with_priority {
+  struct decoder_with_priority
+  {
     const heif_decoder_plugin* plugin;
     int priority;
   };
@@ -1902,13 +2024,13 @@ int heif_get_decoder_descriptors(enum heif_compression_format format_filter,
   std::vector<decoder_with_priority> plugins;
   std::vector<heif_compression_format> formats;
   if (format_filter == heif_compression_undefined) {
-    formats = { heif_compression_HEVC, heif_compression_AV1, heif_compression_VVC };
+    formats = {heif_compression_HEVC, heif_compression_AV1, heif_compression_JPEG, heif_compression_JPEG2000, heif_compression_VVC};
   }
   else {
     formats.emplace_back(format_filter);
   }
 
-  for (const auto* plugin : s_decoder_plugins) {
+  for (const auto* plugin : get_decoder_plugins()) {
     for (auto& format : formats) {
       int priority = plugin->does_support_format(format);
       if (priority) {
@@ -1919,17 +2041,17 @@ int heif_get_decoder_descriptors(enum heif_compression_format format_filter,
   }
 
   if (out_decoders == nullptr) {
-    return (int)plugins.size();
+    return (int) plugins.size();
   }
 
   std::sort(plugins.begin(), plugins.end(), [](const decoder_with_priority& a, const decoder_with_priority& b) {
     return a.priority > b.priority;
   });
 
-  int nDecodersReturned = std::min(count, (int)plugins.size());
+  int nDecodersReturned = std::min(count, (int) plugins.size());
 
-  for (int i=0;i<nDecodersReturned;i++) {
-    out_decoders[i] = (heif_decoder_descriptor*)(plugins[i].plugin);
+  for (int i = 0; i < nDecodersReturned; i++) {
+    out_decoders[i] = (heif_decoder_descriptor*) (plugins[i].plugin);
   }
 
   return nDecodersReturned;
@@ -1938,14 +2060,14 @@ int heif_get_decoder_descriptors(enum heif_compression_format format_filter,
 
 const char* heif_decoder_descriptor_get_name(const struct heif_decoder_descriptor* descriptor)
 {
-  auto decoder = (heif_decoder_plugin*)descriptor;
+  auto decoder = (heif_decoder_plugin*) descriptor;
   return decoder->get_plugin_name();
 }
 
 
 const char* heif_decoder_descriptor_get_id_name(const struct heif_decoder_descriptor* descriptor)
 {
-  auto decoder = (heif_decoder_plugin*)descriptor;
+  auto decoder = (heif_decoder_plugin*) descriptor;
   if (decoder->plugin_api_version < 3) {
     return nullptr;
   }
@@ -2013,14 +2135,14 @@ struct heif_error heif_context_get_encoder(struct heif_context* context,
 
 int heif_have_decoder_for_format(enum heif_compression_format format)
 {
-  auto plugin = heif::get_decoder(format, nullptr);
+  auto plugin = get_decoder(format, nullptr);
   return plugin != nullptr;
 }
 
 
 int heif_have_encoder_for_format(enum heif_compression_format format)
 {
-  auto plugin = heif::get_encoder(format);
+  auto plugin = get_encoder(format);
   return plugin != nullptr;
 }
 
@@ -2107,8 +2229,7 @@ struct heif_error heif_encoder_set_logging_level(struct heif_encoder* encoder, i
     return encoder->plugin->set_parameter_logging_level(encoder->encoder, level);
   }
 
-  struct heif_error err = {heif_error_Ok, heif_suberror_Unspecified, kSuccess};
-  return err;
+  return heif_error_success;
 }
 
 
@@ -2208,7 +2329,7 @@ heif_encoder_parameter_get_valid_integer_range(const struct heif_encoder_paramet
     *have_minimum_maximum = param->integer.have_minimum_maximum;
   }
 
-  return error_Ok;
+  return heif_error_success;
 }
 
 LIBHEIF_API
@@ -2256,7 +2377,7 @@ struct heif_error heif_encoder_parameter_get_valid_integer_values(const struct h
     *num_valid_values = param->integer.num_valid_values;
   }
 
-  return error_Ok;
+  return heif_error_success;
 }
 
 
@@ -2272,7 +2393,7 @@ heif_encoder_parameter_get_valid_string_values(const struct heif_encoder_paramet
     *out_stringarray = param->string.valid_values;
   }
 
-  return error_Ok;
+  return heif_error_success;
 }
 
 struct heif_error heif_encoder_parameter_integer_valid_range(struct heif_encoder* encoder,
@@ -2395,7 +2516,7 @@ struct heif_error heif_encoder_set_parameter(struct heif_encoder* encoder,
           break;
       }
 
-      return error_Ok;
+      return heif_error_success;
     }
   }
 
@@ -2448,7 +2569,7 @@ struct heif_error heif_encoder_get_parameter(struct heif_encoder* encoder,
           break;
       }
 
-      return error_Ok;
+      return heif_error_success;
     }
   }
 
@@ -2479,23 +2600,30 @@ int heif_encoder_has_default(struct heif_encoder* encoder,
 
 static void set_default_options(heif_encoding_options& options)
 {
-  options.version = 5;
+  options.version = 6;
 
   options.save_alpha_channel = true;
-  options.macOS_compatibility_workaround = true;
+  options.macOS_compatibility_workaround = false;
   options.save_two_colr_boxes_when_ICC_and_nclx_available = false;
   options.output_nclx_profile = nullptr;
   options.macOS_compatibility_workaround_no_nclx_profile = true;
   options.image_orientation = heif_orientation_normal;
+
+  options.color_conversion_options.version = 1;
+  options.color_conversion_options.preferred_chroma_downsampling_algorithm = heif_chroma_downsampling_average;
+  options.color_conversion_options.preferred_chroma_upsampling_algorithm = heif_chroma_upsampling_bilinear;
+  options.color_conversion_options.only_use_preferred_chroma_algorithm = false;
 }
 
 static void copy_options(heif_encoding_options& options, const heif_encoding_options& input_options)
 {
-  set_default_options(options);
-
   switch (input_options.version) {
+    case 6:
+      options.color_conversion_options = input_options.color_conversion_options;
+      // fallthrough
     case 5:
       options.image_orientation = input_options.image_orientation;
+      // fallthrough
     case 4:
       options.output_nclx_profile = input_options.output_nclx_profile;
       options.macOS_compatibility_workaround_no_nclx_profile = input_options.macOS_compatibility_workaround_no_nclx_profile;
@@ -2540,10 +2668,8 @@ struct heif_error heif_context_encode_image(struct heif_context* ctx,
 
   heif_encoding_options options;
   heif_color_profile_nclx nclx;
-  if (input_options == nullptr) {
-    set_default_options(options);
-  }
-  else {
+  set_default_options(options);
+  if (input_options) {
     copy_options(options, *input_options);
 
     if (options.output_nclx_profile == nullptr) {
@@ -2551,9 +2677,9 @@ struct heif_error heif_context_encode_image(struct heif_context* ctx,
       if (input_nclx) {
         options.output_nclx_profile = &nclx;
         nclx.version = 1;
-        nclx.color_primaries = (enum heif_color_primaries)input_nclx->get_colour_primaries();
-        nclx.transfer_characteristics = (enum heif_transfer_characteristics)input_nclx->get_transfer_characteristics();
-        nclx.matrix_coefficients = (enum heif_matrix_coefficients)input_nclx->get_matrix_coefficients();
+        nclx.color_primaries = (enum heif_color_primaries) input_nclx->get_colour_primaries();
+        nclx.transfer_characteristics = (enum heif_transfer_characteristics) input_nclx->get_transfer_characteristics();
+        nclx.matrix_coefficients = (enum heif_matrix_coefficients) input_nclx->get_matrix_coefficients();
         nclx.full_range_flag = input_nclx->get_full_range_flag();
       }
     }
@@ -2565,7 +2691,7 @@ struct heif_error heif_context_encode_image(struct heif_context* ctx,
 
   error = ctx->context->encode_image(input_image->image,
                                      encoder,
-                                     &options,
+                                     options,
                                      heif_image_input_class_normal,
                                      image);
   if (error != Error::Ok) {
@@ -2584,7 +2710,7 @@ struct heif_error heif_context_encode_image(struct heif_context* ctx,
     (*out_image_handle)->context = ctx->context;
   }
 
-  return error_Ok;
+  return heif_error_success;
 }
 
 
@@ -2601,16 +2727,17 @@ struct heif_error heif_context_encode_thumbnail(struct heif_context* ctx,
                                                 const struct heif_image* image,
                                                 const struct heif_image_handle* image_handle,
                                                 struct heif_encoder* encoder,
-                                                const struct heif_encoding_options* options,
+                                                const struct heif_encoding_options* input_options,
                                                 int bbox_size,
                                                 struct heif_image_handle** out_image_handle)
 {
   std::shared_ptr<HeifContext::Image> thumbnail_image;
 
-  heif_encoding_options default_options;
-  if (options == nullptr) {
-    set_default_options(default_options);
-    options = &default_options;
+  heif_encoding_options options;
+  set_default_options(options);
+
+  if (input_options != nullptr) {
+    copy_options(options, *input_options);
   }
 
   Error error = ctx->context->encode_thumbnail(image->image,
@@ -2645,7 +2772,7 @@ struct heif_error heif_context_encode_thumbnail(struct heif_context* ctx,
     }
   }
 
-  return error_Ok;
+  return heif_error_success;
 }
 
 
@@ -2654,7 +2781,7 @@ struct heif_error heif_context_set_primary_image(struct heif_context* ctx,
 {
   ctx->context->set_primary_image(image_handle->image);
 
-  return error_Ok;
+  return heif_error_success;
 }
 
 
@@ -2667,7 +2794,7 @@ struct heif_error heif_context_add_exif_metadata(struct heif_context* ctx,
     return error.error_struct(ctx->context.get());
   }
   else {
-    return error_Ok;
+    return heif_error_success;
   }
 }
 
@@ -2691,7 +2818,7 @@ struct heif_error heif_context_add_XMP_metadata2(struct heif_context* ctx,
     return error.error_struct(ctx->context.get());
   }
   else {
-    return error_Ok;
+    return heif_error_success;
   }
 }
 
@@ -2707,7 +2834,7 @@ struct heif_error heif_context_add_generic_metadata(struct heif_context* ctx,
     return error.error_struct(ctx->context.get());
   }
   else {
-    return error_Ok;
+    return heif_error_success;
   }
 }
 
